@@ -7,31 +7,34 @@
 #include <strings.h>
 
 #define liste_bombe_size (247)
+#define INFINITE_SHIELD 1000000
+#define COUNTDOWN_APOCALYPSE 64
 
 int playerGrid[NUMBER_OF_CELLS];
+bool humanPlayer[NUMBER_OF_CELLS]; // is there an human player in a cell
+int victoriesGrid[NUMBER_OF_CELLS]; // biggest number of victories for players in cell
 int lastPlayerGridUpdate=0;
 
-static int getAdderX(int player) {
-	return GETXPIXELSTOCENTEROFCELL(player)*framesToCrossACell(player)/CELLPIXELSSIZE;
-}
-static int getAdderY(int player) {
-	return GETYPIXELSTOCENTEROFCELL(player)*framesToCrossACell(player)/CELLPIXELSSIZE;
-}
-
-
 void inline updatePlayerGrid()
 {
 	if ((!lastPlayerGridUpdate) || (frameNumber()!=lastPlayerGridUpdate))
 	{
-		for (int i=0; i<NUMBER_OF_CELLS; i++)
+		for (int i=0; i<NUMBER_OF_CELLS; i++) {
 			playerGrid[i]=0;
-
+			humanPlayer[i]=false;
+			victoriesGrid[i]=0;
+		}
 		for (int i=0; i<numberOfPlayers(); i++)
 		{
 			if (isAlive(i))
 			{
 				int cell=cellPlayer(i);
 				playerGrid[cell]=teamOfPlayer(i) | playerGrid[cell];
+				if (!isAIActiveForPlayer(i)) {
+					humanPlayer[cell]=true;
+				}
+				int v=victories(i);
+				if (victoriesGrid[cell]<v) victoriesGrid[cell]=v;
 			}
 		}
 		for (int i=numberOfPlayers(); i<nb_dyna; i++)
@@ -39,7 +42,7 @@ void inline updatePlayerGrid()
 			if (isAlive(i))
 			{
 				int cell=cellPlayer(i);
-				playerGrid[cell]=monster | playerGrid[cell];
+				playerGrid[cell]=monster_team | playerGrid[cell];
 			}
 		}
 		lastPlayerGridUpdate=frameNumber();
@@ -49,7 +52,7 @@ void inline updatePlayerGrid()
 bool monsterInCell(int x,int y)
 {
 	updatePlayerGrid();
-	return  (playerGrid[CELLINDEX(x,y)] & monster);
+	return  (playerGrid[CELLINDEX(x,y)] & monster_team);
 }
 
 bool playerInCell(int x,int y)
@@ -68,45 +71,25 @@ bool playerInCell(int player,int x,int y)
 bool playerNotFromMyTeamInCell(int player,int x,int y)
 {
 	updatePlayerGrid();
-	int notMyTeamMask=(~teamOfPlayer(player)) & (~monster);
+	int notMyTeamMask=(~teamOfPlayer(player)) & (~monster_team);
 	return notMyTeamMask & playerGrid[CELLINDEX(x,y)];
 }
 
-bool enemyAroundCell(int player,const travelCostGrid& travelGrid,int x,int y) {
+bool enemyAroundCell(int player,int x,int y) {
 	updatePlayerGrid();
 	if ((x>=grid_size_x-1) || (!x) || (y>=grid_size_y-1) || (!y)) return false;
 	int notMyTeamMask=~teamOfPlayer(player);
 	int cell=CELLINDEX(x,y);
-
 	int closeMonsterMask=0;
-
-	if (travelGrid.canWalk(cell)) {
-		closeMonsterMask=closeMonsterMask | playerGrid[cell];
-	}
-	if (travelGrid.canWalk(cell+1)) {
-		closeMonsterMask=closeMonsterMask | playerGrid[cell+1];
-	}
-	if (travelGrid.canWalk(cell-1)) {
-		closeMonsterMask=closeMonsterMask | playerGrid[cell-1];
-	}
-	if (travelGrid.canWalk(cell-grid_size_x)) {
-		closeMonsterMask=closeMonsterMask | playerGrid[cell-grid_size_x];
-	}
-	if (travelGrid.canWalk(cell-grid_size_x-1)) {
-		closeMonsterMask=closeMonsterMask | playerGrid[cell-grid_size_x-1];
-	}
-	if (travelGrid.canWalk(cell-grid_size_x+1)) {
-		closeMonsterMask=closeMonsterMask | playerGrid[cell-grid_size_x+1];
-	}
-	if (travelGrid.canWalk(cell+grid_size_x)) {
-		closeMonsterMask=closeMonsterMask | playerGrid[cell+grid_size_x];
-	}
-	if (travelGrid.canWalk(cell+grid_size_x-1)) {
-		closeMonsterMask=closeMonsterMask | playerGrid[cell+grid_size_x-1];
-	}
-	if (travelGrid.canWalk(cell+grid_size_x+1)) {
-		closeMonsterMask=closeMonsterMask | playerGrid[cell+grid_size_x+1];
-	}
+	closeMonsterMask=closeMonsterMask | playerGrid[cell];
+	closeMonsterMask=closeMonsterMask | playerGrid[cell+1];
+	closeMonsterMask=closeMonsterMask | playerGrid[cell-1];
+	closeMonsterMask=closeMonsterMask | playerGrid[cell-grid_size_x];
+	closeMonsterMask=closeMonsterMask | playerGrid[cell-grid_size_x-1];
+	closeMonsterMask=closeMonsterMask | playerGrid[cell-grid_size_x+1];
+	closeMonsterMask=closeMonsterMask | playerGrid[cell+grid_size_x];
+	closeMonsterMask=closeMonsterMask | playerGrid[cell+grid_size_x-1];
+	closeMonsterMask=closeMonsterMask | playerGrid[cell+grid_size_x+1];
 	return notMyTeamMask & closeMonsterMask;
 }
 
@@ -114,24 +97,22 @@ bool shouldPlayerFearCulDeSac(int player,int x,int y) {
 	enum Bonus bonus=bonusInCell(x,y);
 	if ((bonus==bonus_heart) || (bonus==bonus_egg) || (bonus==bonus_bulletproofjacket)) return false;
 	if (invincibility(player)) return false; //TOFIX
-	if (hasKangaroo(player)) return false;
 	return true;
 }
 
 bool isCellCulDeSac(int x,int y) {
 	if ((x>=grid_size_x-1) || (!x) || (y>=grid_size_y-1) || (!y)) return false;
 	int i=0;
-	if (!(somethingThatIsNoTABombAndThatWouldStopPlayer(x,y+1) || bombInCell(x,y+1))) i++;
+	if (!(brickOrSkullBonus(x,y+1) || bombInCell(x,y+1))) i++;
 	if (i>1) return false;
-	if (!(somethingThatIsNoTABombAndThatWouldStopPlayer(x,y-1) || bombInCell(x,y-1))) i++;
+	if (!(brickOrSkullBonus(x,y-1) || bombInCell(x,y-1))) i++;
 	if (i>1) return false;
-	if (!(somethingThatIsNoTABombAndThatWouldStopPlayer(x-1,y) || bombInCell(x-1,y))) i++;
+	if (!(brickOrSkullBonus(x-1,y) || bombInCell(x-1,y))) i++;
 	if (i>1) return false;
-	if (!(somethingThatIsNoTABombAndThatWouldStopPlayer(x+1,y) || bombInCell(x+1,y))) i++;
+	if (!(brickOrSkullBonus(x+1,y) || bombInCell(x+1,y))) i++;
 	if (i>1) return false;
 	return true;
 }
-
 // TOFIX: naive...
 static int distance(int x,int y,int xP,int yP) {
 	return abs(x-xP)+abs(y-yP);
@@ -179,11 +160,11 @@ void iterateOnBombs(FunctionWithBombInfo f)
 	assert(index<liste_bombe_size);
 }
 
-void drawBombFlames(int cell, int flameSize, FunctionWithThreeInts f)
+void drawBombFlames(int player, int cell, int flameSize, FunctionWithFlameDrawingHelpfulData f,uint32_t flameGrid[grid_size_x][grid_size_y], bool dangerGrid[grid_size_x][grid_size_y],int &score)
 {
 	int x=CELLX(cell);
 	int y=CELLY(cell);
-	f(x,y,0);
+	f(player,x,y,0,flameGrid,dangerGrid,score);
 	int xx=x;
 	int yy=y;
 	int fs=flameSize;
@@ -191,7 +172,7 @@ void drawBombFlames(int cell, int flameSize, FunctionWithThreeInts f)
 	{
 		xx--;
 		fs--;
-		f(xx,yy,flameSize-fs);
+		f(player,xx,yy,flameSize-fs,flameGrid,dangerGrid,score);
 		if (somethingThatWouldStopFlame(xx,yy))
 			break;
 	}
@@ -202,7 +183,7 @@ void drawBombFlames(int cell, int flameSize, FunctionWithThreeInts f)
 	{
 		yy--;
 		fs--;
-		f(xx,yy,flameSize-fs);
+		f(player,xx,yy,flameSize-fs,flameGrid,dangerGrid,score);
 		if (somethingThatWouldStopFlame(xx,yy))
 			break;
 	}
@@ -213,7 +194,7 @@ void drawBombFlames(int cell, int flameSize, FunctionWithThreeInts f)
 	{
 		xx++;
 		fs--;
-		f(xx,yy,flameSize-fs);
+		f(player,xx,yy,flameSize-fs,flameGrid,dangerGrid,score);
 		if (somethingThatWouldStopFlame(xx,yy))
 			break;
 	}
@@ -224,7 +205,7 @@ void drawBombFlames(int cell, int flameSize, FunctionWithThreeInts f)
 	{
 		yy++;
 		fs--;
-		f(xx,yy,flameSize-fs);
+		f(player,xx,yy,flameSize-fs,flameGrid,dangerGrid,score);
 		if (somethingThatWouldStopFlame(xx,yy))
 			break;
 	}
@@ -385,7 +366,7 @@ enum Button howToGo(int player, int toX,int toY,const travelCostGrid& travelGrid
 		return howToGo(player, toXChosen,toYChosen,travelGrid,shouldJump);
 	}
 }
-static bool canPlayerJump(int player,int x,int y,int inVbls, int fromDirection,const uint32_t flameGrid[grid_size_x][grid_size_y]) {
+static bool canPlayerJump(int player,int x,int y,int inVbls, int fromDirection,const uint32_t flameGrid[grid_size_x][grid_size_y],const bool dangerGrid[grid_size_x][grid_size_y]) {
 	if (hasKangaroo(player)) {
 		int x2=x;
 		int y2=y;
@@ -411,6 +392,9 @@ static bool canPlayerJump(int player,int x,int y,int inVbls, int fromDirection,c
 			assert(0);
 			break;
 		}
+		if (dangerGrid[x2][y2]) {
+			return false;
+		}
 		if (somethingThatIsNoTABombAndThatWouldStopPlayer(x2,y2)) {
 			return false;
 		}
@@ -427,20 +411,29 @@ static bool canPlayerJump(int player,int x,int y,int inVbls, int fromDirection,c
 }
 
 
-static bool canPlayerWalk(int player,int x,int y,int inVbls, int fromDirection,const uint32_t flameGrid[grid_size_x][grid_size_y]) {
+static bool canPlayerWalk(int player, int invincibility, int x,int y,int inVbls, int fromDirection,const uint32_t flameGrid[grid_size_x][grid_size_y],const bool dangerGrid[grid_size_x][grid_size_y]) {
 	int danger=flameGrid[x][y]-inVbls;
-	int shield=invincibility(player)-inVbls;
+	int shield=invincibility-inVbls;
+
+	if (dangerGrid[x][y]) {
+		return false;
+	}
 
 	if ((danger>0) && (danger<=FLAME_DURATION) && (shield<=0) && (bonusInCell(x,y)!=bonus_bulletproofjacket)) {
 		return false;
 	}
 
-	if (somethingThatIsNoTABombAndThatWouldStopPlayer(x,y)) {
+	if (brickOrSkullBonus(x,y)) {
+		return false;
+	}
+
+	if ((monsterInCell(x,y)) && (shield<=0)) {
 		return false;
 	}
 
 	if (bombInCell(x,y)) {
 		if (hasPush(player)) {
+			if ((playerInCell(x,y) || monsterInCell(x,y))) return false;
 			int x2=x;
 			int y2=y;
 			switch (fromDirection) {
@@ -477,22 +470,38 @@ static bool canPlayerWalk(int player,int x,int y,int inVbls, int fromDirection,c
 
 
 // fromDistance is the distance from the bomb center
-static int scoreForBombingCell(int player,int x,int y,int fromDistance,int flameSize)
+static int scoreForBombingCell(int player,int x,int y,int fromDistance,int flameSize, bool ignoreBricks)
 {
 	int result=0;
 
-	if (playerNotFromMyTeamInCell(player,x,y))
+	if (playerNotFromMyTeamInCell(player,x,y)) {
+		if (humanPlayer[CELLINDEX(x,y)]) result++; // focus on humans
+		result+=victoriesGrid[CELLINDEX(x,y)]; // focus on players with more victories
 		result+=3;
+	}
 
 	if (monsterInCell(x,y))
 	{
 		int monsterScore=4*(fromDistance+1);
 		result+=monsterScore;
 	}
+
+	if (ignoreBricks) return result;
+
+	if (bombInCell(x,y)) {
+		result+=2;
+	}
+
 	enum Bonus bonus=bonusInCell(x,y);
 	if (bonus!=no_bonus) {
 		if (bonusPlayerWouldLike(player,bonus)==false) result+=2;
 	}
+
+	// dont care about blowing bricks when 1 bomb left and monster in the vicinity
+	if ((howManyBombsHasPlayerLeft(player)<2) && canPlayerBeReachedByMonster(player)) {
+		return result;
+	}
+
 	if (mudbrickInCell(x,y)) {
 		result++;
 		if ((mudbrickInCell(x+1,y)) || (brickInCell(x+1,y))) result++;
@@ -503,15 +512,9 @@ static int scoreForBombingCell(int player,int x,int y,int fromDistance,int flame
 	return result;
 }
 
-// used by increaseScoreAndUpdateGrid and updateBestExplosionGrid
-static uint32_t grid[grid_size_x][grid_size_y];
-static int score;
-static int playerSave;
-static int flame;
-
-static void updateScoreFunctionFunctionWithThreeInts(int x,int y,int distance) {
-	score+=scoreForBombingCell(playerSave,x,y,distance,flame);
-	grid[x][y]=COUNTDOWN_DURATON+FLAME_DURATION;
+static void updateScoreFunctionFunctionWithFlameDrawingHelpfulData(int player, int x,int y,int distance,uint32_t flameGrid[grid_size_x][grid_size_y], bool dangerGrid[grid_size_x][grid_size_y],int &score) {
+	score+=scoreForBombingCell(player,x,y,distance,flameSize(player),false);
+	flameGrid[x][y]=COUNTDOWN_DURATON+FLAME_DURATION;
 }
 
 void updateBestExplosionGrid(int player,
@@ -521,23 +524,25 @@ void updateBestExplosionGrid(int player,
                              const bool dangerGrid[grid_size_x][grid_size_y])
 {
 	// calculate the best place to drop a bomb
-	playerSave=player;
-	flame=flameSize(player);
+	int currentCell=cellPlayer(player);
+
 	for (int j=0; j<grid_size_y; j++)
 	{
 		for (int i=0; i<grid_size_x; i++)
 		{
-			score=0;
-
+			int score=0;
 			if (
-				dangerGrid[i][j]==false
+				(dangerGrid[i][j]==false || ((CELLX(currentCell)==i) && (CELLY(currentCell))==j)) //authorise the current player cell even if it's in the dangerGrid
 				&& travelGrid.canWalk(i,j)
 				&& (flameGrid[i][j]==0 || travelGrid.cost(i,j)>flameGrid[i][j])
+				&& bombInCell(i,j)==false
 				)
 			{
+				uint32_t grid[grid_size_x][grid_size_y];
 				memmove(grid, flameGrid, sizeof(grid));
+				bool unusedDangerGrid[grid_size_x][grid_size_y];
 
-				drawBombFlames(CELLINDEX(i,j),flame,updateScoreFunctionFunctionWithThreeInts);
+				drawBombFlames(player,CELLINDEX(i,j),flameSize(player),updateScoreFunctionFunctionWithFlameDrawingHelpfulData,grid,unusedDangerGrid,score);
 
 				// check that there is still a safe place in the grid:
 				bool foundSafePlace=false;
@@ -561,14 +566,13 @@ void updateBestExplosionGrid(int player,
 	}
 }
 
-#define countDownApocalypse 64
 
 static bool apocalyseDangerForCell(int x,int y)
 {
 	if (isInTheApocalypse())
 	{
 		db danger=m.truc_fin[x+y*grid_size_x_with_padding];
-		if (danger<countDownApocalypse)
+		if (danger<COUNTDOWN_APOCALYPSE)
 			return true;
 	}
 
@@ -576,12 +580,12 @@ static bool apocalyseDangerForCell(int x,int y)
 }
 
 
-static void visitCell(int player, int currentCell,
-                      const uint32_t flameGrid[grid_size_x][grid_size_y],
+static void visitCell(int player, bool ignoreFlames, int currentCell,
+                      const uint32_t flameGrid[grid_size_x][grid_size_y],const bool dangerGrid[grid_size_x][grid_size_y],
                       int adderX,int adderY,int framesPerCell, int direction,travelCostGrid& travelGrid, std::priority_queue<std::pair<int,int> > &queue, bool visited[NUMBER_OF_CELLS]) {
 	int nextCell;
 	uint32_t nextCost=travelGrid.cost(currentCell)+framesPerCell;
-	int adderCell;
+	int adderCell=0;
 	switch (direction) {
 	case button_right:
 		nextCost+=-adderX+abs(adderY);
@@ -608,13 +612,13 @@ static void visitCell(int player, int currentCell,
 	uint32_t nextCost2=nextCost+framesPerCell;
 	if (!visited[nextCell]) {
 
-		if (canPlayerWalk(player,CELLX(nextCell),CELLY(nextCell),nextCost,direction,flameGrid))
+		if (canPlayerWalk(player,ignoreFlames ? INFINITE_SHIELD : invincibility(player)-framesPerCell,CELLX(nextCell),CELLY(nextCell),nextCost,direction,flameGrid,dangerGrid))
 		{
 			if (nextCost<travelGrid.cost(nextCell)) {
 				travelGrid.setWalkingCost(nextCell,nextCost);
 				queue.push(std::pair <int,int>(-nextCost, nextCell));
 			}
-		} else if (canPlayerJump(player,CELLX(nextCell),CELLY(nextCell),nextCost2,direction,flameGrid)) {
+		} else if (!ignoreFlames && (canPlayerJump(player,CELLX(nextCell),CELLY(nextCell),nextCost2,direction,flameGrid,dangerGrid))) {
 			if ((travelGrid.jumpingCost(nextCell,direction)>=nextCost) && (travelGrid.cost(nextCell2)>=nextCost2)) {
 				travelGrid.setJumpingCost(nextCell,nextCost,direction);
 				travelGrid.setWalkingCost(nextCell2,nextCost2);
@@ -624,9 +628,10 @@ static void visitCell(int player, int currentCell,
 	}
 }
 
-void updateTravelGrid(int player,
+void updateTravelGrid(int player, bool ignoreFlames,
                       travelCostGrid& travelGrid,
-                      const uint32_t flameGrid[grid_size_x][grid_size_y])
+                      const uint32_t flameGrid[grid_size_x][grid_size_y],
+                      const bool dangerGrid[grid_size_x][grid_size_y])
 {
 	bool visited[NUMBER_OF_CELLS];
 	std::priority_queue<std::pair<int,int> > queue;
@@ -638,13 +643,13 @@ void updateTravelGrid(int player,
 	int currentCell=playerCell;
 	int framesPerCell=framesToCrossACell(player);
 
-	travelGrid.setWalkingCost(currentCell,0);
-
 	for (int i=0; i<NUMBER_OF_CELLS; i++) {
 		visited[i]=false;
 		if (CELLX(i)==0 || CELLX(i)==grid_size_x-1 || CELLY(i)==0 || CELLY(i)==grid_size_y-1 ) visited[i]=true;
 	}
+
 	travelGrid.setWalkingCost(currentCell,0);
+
 	queue.push(std::pair <int,int> (0, currentCell));
 	while(!queue.empty()) {
 		pair = queue.top();
@@ -652,10 +657,10 @@ void updateTravelGrid(int player,
 		currentCell = pair.second;
 		if (!visited[currentCell]) {
 			visited[currentCell] = true;
-			visitCell(player,currentCell,flameGrid,adderX, adderY, framesPerCell,button_right,travelGrid,queue,visited);
-			visitCell(player,currentCell,flameGrid,adderX, adderY, framesPerCell,button_left,travelGrid,queue,visited);
-			visitCell(player,currentCell,flameGrid,adderX, adderY, framesPerCell,button_up,travelGrid,queue,visited);
-			visitCell(player,currentCell,flameGrid,adderX, adderY, framesPerCell,button_down,travelGrid,queue,visited);
+			visitCell(player,ignoreFlames,currentCell,flameGrid,dangerGrid,adderX, adderY, framesPerCell,button_right,travelGrid,queue,visited);
+			visitCell(player,ignoreFlames,currentCell,flameGrid,dangerGrid,adderX, adderY, framesPerCell,button_left,travelGrid,queue,visited);
+			visitCell(player,ignoreFlames,currentCell,flameGrid,dangerGrid,adderX, adderY, framesPerCell,button_up,travelGrid,queue,visited);
+			visitCell(player,ignoreFlames,currentCell,flameGrid,dangerGrid,adderX, adderY, framesPerCell,button_down,travelGrid,queue,visited);
 			adderX=0;
 			adderY=0;
 		}
@@ -663,52 +668,100 @@ void updateTravelGrid(int player,
 	travelGrid.setWalkingCost(playerCell,abs(getAdderX(player))+abs(getAdderY(player)));
 }
 
-
-// used by increaseScoreAndUpdateGrid and updateBestExplosionGrid
-static bool dangerGrid_save[grid_size_x][grid_size_y];
-static uint32_t flameGrid_save[grid_size_x][grid_size_y];
-static uint32_t countDown;
-
-static void updateFlameAndDangerGridsFunctionFunctionWithThreeInts(int x,int y,int distance) {
-	flameGrid_save[x][y]=flameGrid_save[x][y] ? std::min(flameGrid_save[x][y],countDown) : countDown;
-	dangerGrid_save[x][y]=true;
+static void updateFlameAndDangerGridsFunctionFunctionWithThreeInts(int player, int x,int y,int distance,uint32_t flameGrid[grid_size_x][grid_size_y], bool dangerGrid[grid_size_x][grid_size_y],int &countDown) {
+	flameGrid[x][y]=flameGrid[x][y] ? std::min(flameGrid[x][y],uint32_t(countDown)) : countDown;
+	if (!countDown) {
+		dangerGrid[x][y]=true;
+	}
 }
 static std::vector < struct bombInfo * > vec;
 static void addBombsIntoVector(struct bombInfo * bomb)
 {
 	vec.push_back(bomb);
 }
+static void updateScoreFunctionFunction(int player, int x,int y,int distance,uint32_t flameGrid[grid_size_x][grid_size_y], bool dangerGrid[grid_size_x][grid_size_y],int &score) {
+	score+=scoreForBombingCell(player,x,y,distance,flameSize(player),true);
+	dangerGrid[x][y]=true;
+}
+int calculateScoreForActivatingRemote(int player) {
+	if (!hasRemote(player)) return 0;
+	int score=0;
+	uint32_t unusedFlameGrid[grid_size_x][grid_size_y];
+	bool bombedGrid[grid_size_x][grid_size_y];
 
-void updateFlameAndDangerGridsWithBombs(int player,uint32_t flameGrid[grid_size_x][grid_size_y],bool dangerGrid[grid_size_x][grid_size_y])
-{
 	for (int j=0; j<grid_size_y; j++)
 	{
 		for (int i=0; i<grid_size_x; i++)
 		{
-			flameGrid_save[i][j]=0;
-			dangerGrid_save[i][j]=false;
+			bombedGrid[i][j]=false;
+		}
+	}
+
+	vec.clear();
+	iterateOnBombs(addBombsIntoVector);
+	for (std::vector<struct bombInfo *>::iterator it = vec.begin(); it != vec.end(); ++it) {
+		struct bombInfo * bomb=*it;
+		int i=bomb->x();
+		int j=bomb->y();
+		if (bomb->getPlayer()==player) {
+			drawBombFlames(player,CELLINDEX(i,j),flameSize(player),updateScoreFunctionFunction,unusedFlameGrid,bombedGrid,score);
+		}
+	}
+// chain effect
+	for (int z=0; z<4; z++) {
+		for (std::vector<struct bombInfo *>::iterator it = vec.begin(); it != vec.end(); ++it) {
+			struct bombInfo * bomb=*it;
+			int i=bomb->x();
+			int j=bomb->y();
+			if ((bomb->getPlayer()!=player) && (bombedGrid[i][j]==true)) {
+				drawBombFlames(player,CELLINDEX(i,j),flameSize(bomb->getPlayer()),updateScoreFunctionFunction,unusedFlameGrid,bombedGrid,score);
+			}
+		}
+	}
+	return score;
+}
+
+
+void updateFlameAndDangerGridsWithBombs(int player,uint32_t flameGrid[grid_size_x][grid_size_y],bool dangerGrid[grid_size_x][grid_size_y])
+{
+	struct bombInfo possibleShieldRemoteBombsUnderPlayer[nb_dyna];
+	for (int j=0; j<grid_size_y; j++)
+	{
+		for (int i=0; i<grid_size_x; i++)
+		{
+			flameGrid[i][j]=0;
+			dangerGrid[i][j]=false;
 		}
 	}
 	vec.clear();
 	iterateOnBombs(addBombsIntoVector);
-
+// add "virtual bombs" under other players that have remote + shields
+	for (int i=0; i<numberOfPlayers(); i++)
+	{
+		if (player!=i && isAlive(i) && hasRemote(i) && invincibility(i))
+		{
+			possibleShieldRemoteBombsUnderPlayer[i].remote=1;
+			possibleShieldRemoteBombsUnderPlayer[i].cell(cellPlayer(i));
+			possibleShieldRemoteBombsUnderPlayer[i].countDown=0;
+			possibleShieldRemoteBombsUnderPlayer[i].flameSize=flameSize(i);
+			vec.push_back(&possibleShieldRemoteBombsUnderPlayer[i]);
+		}
+	}
+//
 	for (std::vector<struct bombInfo *>::iterator it = vec.begin(); it != vec.end(); ++it) {
 		struct bombInfo * bomb=*it;
 
-		countDown=(int)
-		           bomb->countDown+FLAME_DURATION;
+		int countDown=(int)
+		               bomb->countDown+FLAME_DURATION;
 		if (bomb->remote)
 			countDown=0;
 		int i=bomb->x();
 		int j=bomb->y();
-		if (flameGrid_save[i][j])
-			countDown=std::min(flameGrid_save[i][j],countDown); //this enable bomb explosions chains
+		if (flameGrid[i][j])
+			countDown=std::min(flameGrid[i][j],uint32_t(countDown)); //this enable bomb explosions chains
 
-		drawBombFlames(CELLINDEX(bomb->x(),bomb->y()),bomb->flameSize,updateFlameAndDangerGridsFunctionFunctionWithThreeInts);
+		drawBombFlames(player,CELLINDEX(bomb->x(),bomb->y()),bomb->flameSize,updateFlameAndDangerGridsFunctionFunctionWithThreeInts,flameGrid,dangerGrid,countDown);
 	}
-	memmove(dangerGrid, dangerGrid_save,  sizeof(dangerGrid_save));
-	memmove(flameGrid, flameGrid_save,  sizeof(flameGrid_save));
-
 	for (int j=0; j<grid_size_y; j++)
 	{
 		for (int i=0; i<grid_size_x; i++)
@@ -719,9 +772,15 @@ void updateFlameAndDangerGridsWithBombs(int player,uint32_t flameGrid[grid_size_
 	}
 }
 
-void updateDangerGridWithMonstersAndCulDeSacs(int player, const travelCostGrid& travelGrid,bool dangerGrid[grid_size_x][grid_size_y])
+void updateDangerGridWithMonstersSickPlayersAndCulDeSacs(int player, bool dangerGrid[grid_size_x][grid_size_y])
 {
-
+	for (int i=0; i<numberOfPlayers(); i++) {
+		if (isAlive(i) && hasAnyDisease(i) && player!=i)
+		{
+			int cell=cellPlayer(i);
+			dangerGrid[CELLX(cell)][CELLY(cell)]=true;
+		}
+	}
 	for (int i=numberOfPlayers(); i<nb_dyna; i++)
 	{
 		if (isAlive(i))
@@ -741,7 +800,7 @@ void updateDangerGridWithMonstersAndCulDeSacs(int player, const travelCostGrid&
 				if (invincibility(player)>FLAME_DURATION) {
 					dangerGrid[i][j]=false;
 				} else {
-					if ((enemyAroundCell(player,travelGrid, i,j)) && (isCellCulDeSac(i,j)) && shouldPlayerFearCulDeSac(player,i,j)) {
+					if ((enemyAroundCell(player, i,j)) && (isCellCulDeSac(i,j)) && shouldPlayerFearCulDeSac(player,i,j)) {
 						dangerGrid[i][j]=true;
 					}
 				}
@@ -764,6 +823,116 @@ void updateMonsterIsComingGrid(bool monsterIsComingGrid[NUMBER_OF_CELLS]) {
 		}
 	}
 }
+void updateDangerGridWithMonster4CellsTerritories(bool dangerGrid[grid_size_x][grid_size_y]) {
+	static int frame;
+	static bool init = true;
+	static uint32_t noFlameGrid[grid_size_x][grid_size_y];
+	static bool noDangerGrid[grid_size_x][grid_size_y];
+	static bool resultDangerGrid[grid_size_x][grid_size_y];
+	if (init) {
+		for (int j=0; j<grid_size_y; j++) {
+			for (int i=0; i<grid_size_x; i++) {
+				noFlameGrid[i][j]=0;
+				noDangerGrid[i][j]=false;
+			}
+		}
+		init=false;
+		frame=frameNumber()-1;
+	}
+
+	if (frame!=frameNumber()) {
+
+		for (int j=0; j<grid_size_y; j++) {
+			for (int i=0; i<grid_size_x; i++) {
+				resultDangerGrid[i][j]=false;
+			}
+		}
+
+
+		for (int monsterIndex=numberOfPlayers(); monsterIndex<nb_dyna; monsterIndex++)
+		{
+			if (isAlive(monsterIndex))
+			{
+				int accessibleCells=0;
+				travelCostGrid travelGrid;
+				updateTravelGrid(monsterIndex,true,travelGrid,noFlameGrid,noDangerGrid);
+				for (int j=0; j<grid_size_y; j++) {
+					for (int i=0; i<grid_size_x; i++) {
+						if (travelGrid.canWalk(i,j)) {
+							accessibleCells++;
+						}
+					}
+				}
+				if (accessibleCells<=4) {
+					for (int j=0; j<grid_size_y; j++) {
+						for (int i=0; i<grid_size_x; i++) {
+							if (travelGrid.canWalk(i,j)) {
+								resultDangerGrid[i][j]=true;
+							}
+						}
+					}
+				}
+			}
+		}
+	}
+
+// update ...
+	for (int j=0; j<grid_size_y; j++) {
+		for (int i=0; i<grid_size_x; i++) {
+			if (resultDangerGrid[i][j]) {
+				dangerGrid[i][j]=true;
+			}
+		}
+	}
+	frame=frameNumber();
+}
+
+bool canPlayerBeReachedByMonster(int player) {
+	static bool result[nb_dyna];
+	static int frame[nb_dyna];
+	static bool init = true;
+	static uint32_t noFlameGrid[grid_size_x][grid_size_y];
+	static bool noDangerGrid[grid_size_x][grid_size_y];
+
+	if (isAlive(player)==false) {
+		return false;
+	}
+
+	travelCostGrid travelGrid;
+	if (init) {
+		for (int j=0; j<grid_size_y; j++) {
+			for (int i=0; i<grid_size_x; i++) {
+				noFlameGrid[i][j]=0;
+				noDangerGrid[i][j]=false;
+			}
+		}
+		init=false;
+		for (int i=0; i<nb_dyna; i++) {
+			frame[i]=frameNumber()-1;
+			result[i]=false;
+		}
+	}
+
+	if (frame[player]!=frameNumber()) {
+		frame[player]=frameNumber();
+	} else {
+		return result[player];
+	}
+	updateTravelGrid(player,true,travelGrid,noFlameGrid,noDangerGrid);
+
+	for (int j=0; j<grid_size_y; j++) {
+		for (int i=0; i<grid_size_x; i++) {
+			if (monsterInCell(i,j) && travelGrid.canWalk(i,j)) {
+				result[player]=true;
+				return result[player];
+			}
+
+		}
+	}
+	result[player]=false;
+	return result[player];
+}
+
 
 void printCellInfo(int cell,int player)
 {
