@@ -25,6 +25,11 @@ static clock_t begin;
 unsigned int nbFrames=0;
 int testAI=0;
 bool slowMode=false;
+#define MAX_TURBO 32
+int turboMode=0;
+int anyButtonPushedMask=0;
+int anyStartButtonPushedMask=0;
+int framesPlayed=0;
 
 void quit(int rc)
 {
@@ -78,9 +83,13 @@ void addJoystick(int index) {
 	}
 }
 
-void UpdateTexture(SDL_Texture *texture)
+void UpdateTexture(SDL_Texture *texture,bool skipRendering)
 {
+
+	#define NB_SCREENS_BLURRING 8
 	static uint32_t matrixPalette[NB_COLORS_PALETTE];
+	static uint32_t previousScreen[WIDTH][HEIGHT][NB_SCREENS_BLURRING];
+	static int numberOfScreenToMelt=1;
 	Uint32 *dst;
 	int row, col;
 	void *pixels;
@@ -99,7 +108,29 @@ void UpdateTexture(SDL_Texture *texture)
 	for (row = 0; row < HEIGHT; ++row) {
 		dst = (Uint32*)((Uint8*)pixels + row * pitch);
 		for (col = 0; col < WIDTH; ++col) {
-			*dst++ = matrixPalette[m.vgaRam[col+row*WIDTH]];
+			uint32_t color=matrixPalette[m.vgaRam[col+row*WIDTH]];
+			previousScreen[col][row][frameNumber()%NB_SCREENS_BLURRING]=color;
+			if (skipRendering) {
+				uint32_t b = 0;
+				uint32_t r = 0;
+				uint32_t g = 0;
+				for (int z=0; z<numberOfScreenToMelt; z++) {
+					uint32_t prevColor=previousScreen[col][row][(NB_SCREENS_BLURRING-z+frameNumber())%NB_SCREENS_BLURRING];
+					b += prevColor&255;
+					r += (prevColor>>8)&255;
+					g += (prevColor>>16)&255;
+				}
+				b = b/(numberOfScreenToMelt);
+				r = r/(numberOfScreenToMelt);
+				g = g/(numberOfScreenToMelt);
+				color=(g<<16) | (r << 8) | b;
+				if (numberOfScreenToMelt<NB_SCREENS_BLURRING-1) {
+					numberOfScreenToMelt+=2;
+				}
+			} else {
+				numberOfScreenToMelt=1;
+			}
+			*dst++ = color;
 		}
 	}
 	SDL_UnlockTexture(texture);
@@ -137,13 +168,19 @@ void  updateKeyboard(Uint8 scancode,int state) {
 		mrboom_update_input(button_a,nb_dyna-2,state,false);
 		break;
 	case SDL_SCANCODE_LCTRL:
+	case SDL_SCANCODE_LGUI:
 		mrboom_update_input(button_b,nb_dyna-2,state,false);
 		break;
 	case SDL_SCANCODE_LSHIFT:
-		mrboom_update_input(button_r,nb_dyna-2,state,false);
+		mrboom_update_input(button_x,nb_dyna-2,state,false);
 		break;
 	case SDL_SCANCODE_SPACE:
 		mrboom_update_input(button_select,nb_dyna-2,state,false);
+		if (state) {
+			anyButtonPushedMask = anyButtonPushedMask | (1<<16);
+		} else {
+			anyButtonPushedMask = anyButtonPushedMask & ~(1<<16);
+		}
 		break;
 	case SDL_SCANCODE_RETURN:
 	case SDL_SCANCODE_KP_ENTER:
@@ -162,19 +199,30 @@ void  updateKeyboard(Uint8 scancode,int state) {
 	case SDL_SCANCODE_RIGHT:
 		mrboom_update_input(button_right,nb_dyna-1,state,false);
 		break;
+	case SDL_SCANCODE_PAGEDOWN:
 	case SDL_SCANCODE_RALT:
 		mrboom_update_input(button_a,nb_dyna-1,state,false);
 		break;
+	case SDL_SCANCODE_END:
 	case SDL_SCANCODE_RCTRL:
+	case SDL_SCANCODE_RGUI:
 		mrboom_update_input(button_b,nb_dyna-1,state,false);
 		break;
+	case SDL_SCANCODE_HOME:
 	case SDL_SCANCODE_RSHIFT:
-		mrboom_update_input(button_r,nb_dyna-1,state,false);
+		mrboom_update_input(button_x,nb_dyna-1,state,false);
 		break;
 	case SDL_SCANCODE_ESCAPE:
-		pressESC();
-		if (isGameActive()==false) {
-			quit(0);
+		if (state) {
+			pressESC();
+			if (inTheMenu()) {
+				done=SDL_TRUE;
+			}
+		}
+		break;
+	case SDL_SCANCODE_PAUSE:
+		if (state) {
+			pauseGameButton();
 		}
 		break;
 	default:
@@ -184,6 +232,17 @@ void  updateKeyboard(Uint8 scancode,int state) {
 }
 
 const int JOYSTICK_DEAD_ZONE = 8000;
+int anyStartButtonPushedCounter=0;
+int beeingPlaying=0;
+uint32_t windowID;
+static const float ASPECT_RATIO=float(WIDTH)/float(HEIGHT);
+
+
+bool resizeDone = false;
+SDL_Rect screen;
+SDL_Window *window;
+int width = 0;
+int height = 0;
 
 void
 loop()
@@ -193,17 +252,60 @@ loop()
 	int xDir = 0;
 	int yDir = 0;
 
+	if (isGameActive()) {
+
+		beeingPlaying++;
+
+		if ((anyStartButtonPushedMask) && (beeingPlaying>20)) {
+			anyStartButtonPushedCounter++;
+			if (anyStartButtonPushedCounter==1) {
+				pauseGameButton();
+			}
+		} else {
+			anyStartButtonPushedCounter=0;
+		}
+
+	} else {
+		beeingPlaying=0;
+	}
+
+
 	while (SDL_PollEvent(&e)) {
 		switch (e.type) {
 
+		case SDL_WINDOWEVENT:
+			if(e.window.windowID == windowID) {
+				switch(e.window.event) {
+				case SDL_WINDOWEVENT_RESIZED: {
+					width = e.window.data1;
+					height = e.window.data2;
+					float aspectRatio = (float)width/(float)height;
+
+					if(aspectRatio != ASPECT_RATIO) {
+						if(aspectRatio > ASPECT_RATIO) {
+							height = (1.f / ASPECT_RATIO) * width;
+						}
+						else {
+							width = ASPECT_RATIO * height;
+						}
+						log_debug("Setting window size to %d, %d, aspect ratio: %f\n",
+						          width, height, (float)width/(float)height);
+					}
+					screen.w = width;
+					screen.h = height;
+					resizeDone = true;
+					break;
+				}
+				}
+			}
+			break;
+
 		case SDL_JOYDEVICEADDED:
 			addJoystick(e.jdevice.which);
 			break;
-
 		case SDL_JOYDEVICEREMOVED:
 			removeJoystick(e.jdevice.which);
 			break;
-
 		case SDL_JOYAXISMOTION:
 			if ((e.jaxis.axis==0) || (e.jaxis.axis==2))
 			{
@@ -223,7 +325,7 @@ loop()
 				}
 			}
 			//Y axis motion
-			else if( e.jaxis.axis == 1 )
+			else if((e.jaxis.axis == 1 ) || (e.jaxis.axis==3))
 			{
 				//Below of dead zone
 				if( e.jaxis.value < -JOYSTICK_DEAD_ZONE )
@@ -240,7 +342,7 @@ loop()
 					yDir = 0;
 				}
 			}
-			if (IFTRACES) log_debug("e.jaxis.axis=%d e.jaxis.value=%d\n",e.jaxis.axis,e.jaxis.value);
+			if ((IFTRACES) && ((xDir) || (yDir))) log_debug("e.jaxis.axis=%d e.jaxis.value=%d player=%d\n",e.jaxis.axis,e.jaxis.value,getPlayerFromJoystickPort(e.jaxis.which));
 			if ((e.jaxis.axis==0) || (e.jaxis.axis==2)) {
 				if (xDir==1) {
 					mrboom_update_input(button_right,getPlayerFromJoystickPort(e.jaxis.which),1,false);
@@ -262,6 +364,70 @@ loop()
 				}
 			}
 			break;
+		case SDL_JOYHATMOTION:
+		{
+			int player=getPlayerFromJoystickPort(e.jhat.which);
+			if (IFTRACES) log_debug("SDL_JOYHATMOTION: .jhat=%d which=%d player=%d\n",e.jhat.value,e.jhat.which,getPlayerFromJoystickPort(e.jhat.which));
+
+			switch(e.jhat.value) {
+
+			case SDL_HAT_CENTERED:
+				mrboom_update_input(button_up,player,0,false);
+				mrboom_update_input(button_down,player,0,false);
+				mrboom_update_input(button_left,player,0,false);
+				mrboom_update_input(button_right,player,0,false);
+				break;
+			case SDL_HAT_UP:
+				mrboom_update_input(button_up,player,1,false);
+				mrboom_update_input(button_down,player,0,false);
+				mrboom_update_input(button_left,player,0,false);
+				mrboom_update_input(button_right,player,0,false);
+				break;
+			case SDL_HAT_DOWN:
+				mrboom_update_input(button_up,player,0,false);
+				mrboom_update_input(button_down,player,1,false);
+				mrboom_update_input(button_left,player,0,false);
+				mrboom_update_input(button_right,player,0,false);
+				break;
+			case SDL_HAT_LEFT:
+				mrboom_update_input(button_up,player,0,false);
+				mrboom_update_input(button_down,player,0,false);
+				mrboom_update_input(button_left,player,1,false);
+				mrboom_update_input(button_right,player,0,false);
+				break;
+			case SDL_HAT_RIGHT:
+				mrboom_update_input(button_up,player,0,false);
+				mrboom_update_input(button_down,player,0,false);
+				mrboom_update_input(button_left,player,0,false);
+				mrboom_update_input(button_right,player,1,false);
+				break;
+			case SDL_HAT_RIGHTDOWN:
+				mrboom_update_input(button_up,player,0,false);
+				mrboom_update_input(button_down,player,1,false);
+				mrboom_update_input(button_left,player,0,false);
+				mrboom_update_input(button_right,player,1,false);
+				break;
+			case SDL_HAT_RIGHTUP:
+				mrboom_update_input(button_up,player,1,false);
+				mrboom_update_input(button_down,player,0,false);
+				mrboom_update_input(button_left,player,0,false);
+				mrboom_update_input(button_right,player,1,false);
+				break;
+			case SDL_HAT_LEFTUP:
+				mrboom_update_input(button_up,player,1,false);
+				mrboom_update_input(button_down,player,0,false);
+				mrboom_update_input(button_left,player,1,false);
+				mrboom_update_input(button_right,player,0,false);
+				break;
+			case SDL_HAT_LEFTDOWN:
+				mrboom_update_input(button_up,player,0,false);
+				mrboom_update_input(button_down,player,1,false);
+				mrboom_update_input(button_left,player,1,false);
+				mrboom_update_input(button_right,player,0,false);
+				break;
+			}
+		}
+		break;
 		case SDL_JOYBUTTONDOWN:
 			if (IFTRACES) log_debug("Joystick %d button %d down\n",
 				                e.jbutton.which, e.jbutton.button);
@@ -285,12 +451,16 @@ loop()
 				mrboom_update_input(button_r,getPlayerFromJoystickPort(e.jbutton.which),1,false);
 				break;
 			case 6:
+			case 8:
 				mrboom_update_input(button_select,getPlayerFromJoystickPort(e.jbutton.which),1,false);
 				break;
 			case 7:
+			case 9:
 				mrboom_update_input(button_start,getPlayerFromJoystickPort(e.jbutton.which),1,false);
+				anyStartButtonPushedMask = anyStartButtonPushedMask | (1<<e.jbutton.which);
 				break;
 			}
+			anyButtonPushedMask = anyButtonPushedMask | (1<<e.jbutton.button);
 			break;
 		case SDL_JOYBUTTONUP:
 			if (IFTRACES) log_debug("Joystick %d button %d up\n",
@@ -315,12 +485,16 @@ loop()
 				mrboom_update_input(button_r,getPlayerFromJoystickPort(e.jbutton.which),0,false);
 				break;
 			case 6:
+			case 8:
 				mrboom_update_input(button_select,getPlayerFromJoystickPort(e.jbutton.which),0,false);
 				break;
 			case 7:
+			case 9:
 				mrboom_update_input(button_start,getPlayerFromJoystickPort(e.jbutton.which),0,false);
+				anyStartButtonPushedMask = anyStartButtonPushedMask  & ~(1<<e.jbutton.which);
 				break;
 			}
+			anyButtonPushedMask = anyButtonPushedMask & ~(1<<e.jbutton.button);
 			break;
 		case SDL_KEYDOWN:
 			updateKeyboard(e.key.keysym.scancode,1);
@@ -338,6 +512,9 @@ loop()
 			break;
 		}
 	}
+	if (someHumanPlayersAlive()) {
+		anyButtonPushedMask=0;
+	}
 
 	mrboom_deal_with_autofire();
 	program();
@@ -353,14 +530,75 @@ loop()
 
 	if (m.executionFinished) done = SDL_TRUE;
 
-	mrboom_play_fx();
+	mrboom_sound();
 
+	bool skipRendering=false;
+	static bool showFrame[MAX_TURBO][MAX_TURBO];
+	static bool calculateShowFrame=true;
+	int counter=0;
+
+#define INT_SCALING 100000
+	if (calculateShowFrame) {
+		for (int i=0; i<MAX_TURBO; i++) {
+			int delta=((MAX_TURBO-i)*INT_SCALING)/MAX_TURBO; // between 100 and 0
+			for (int j=0; j<MAX_TURBO; j++) {
+				counter+=delta;
+				if (counter>=INT_SCALING) {
+					counter-=INT_SCALING;
+					showFrame[i][j]=true;
+				} else {
+					showFrame[i][j]=false;
+				}
+			}
+		}
+		#if 0
+		for (int i=0; i<MAX_TURBO; i++) {
+			for (int j=0; j<MAX_TURBO; j++) {
+				printf("%d ",showFrame[i][j]);
+			}
+			printf("\n");
+		}
+		#endif
+		calculateShowFrame=false;
+	}
+
+	if  (isGameActive() && someHumanPlayersAlive()==false && anyButtonPushedMask) {
+		if (!turboMode) {
+			turboMode=1;
+		}
+		if (!(frameNumber()%32) && (turboMode<MAX_TURBO/2)) {
+			turboMode++;
+		}
+	} else {
+		if (!(frameNumber()%32) && (turboMode)) {
+			turboMode--;
+		}
+	}
+	int phaseShowFrame=turboMode ? turboMode+MAX_TURBO/2 : 0;
+	if (phaseShowFrame>MAX_TURBO-1) {
+		phaseShowFrame=MAX_TURBO-1;
+	}
+	if (showFrame[phaseShowFrame][frameNumber()%MAX_TURBO]) {
+		skipRendering=false;
+	} else {
+		skipRendering=true;
+	}
+	static bool previousSkipRendering=false;
+
+	if  ((isGameActive()==false) || isGamePaused()) {
+		previousSkipRendering=false;
+		turboMode=0;
+	}
 	if (noVGA == SDL_FALSE) {
-		UpdateTexture(texture);
+		UpdateTexture(texture,previousSkipRendering);
+		previousSkipRendering=skipRendering;
+	}
+	if ((noVGA == SDL_FALSE) && skipRendering==false) {
 		SDL_RenderClear(renderer);
 		SDL_RenderCopy(renderer, texture, NULL, NULL);
 		SDL_RenderPresent(renderer);
 	}
+
 #ifdef __EMSCRIPTEN__
 	if (done) {
 		emscripten_cancel_main_loop();
@@ -378,6 +616,38 @@ static void fps() {
 	}
 }
 
+void printKeys() {
+	log_info("Keys   Space - Add a bomberman bot\n");
+	log_info("       Return - Start a game\n");
+	log_info("       ESC - Quit\n");
+	log_info("1st player\n");
+	log_info("       Left - Left\n");
+	log_info("       Right - Right\n");
+	log_info("       Up - Up\n");
+	log_info("       Down - Down\n");
+#ifdef __APPLE__
+	log_info("       Right Ctrl, End or Right Command - Lay bomb\n");
+#else
+	log_info("       Right Ctrl, End or Right GUI - Lay bomb\n");
+#endif
+	log_info("       Right Alt or PageDown - Ignite bomb\n");
+	log_info("       Right Shift or Home - Jump\n");
+	log_info("2nd player\n");
+	log_info("       A - Left\n");
+	log_info("       D - Right\n");
+	log_info("       W - Up\n");
+	log_info("       S - Down\n");
+#ifdef __APPLE__
+	log_info("       Left Ctrl or Left Command - Lay bomb\n");
+#else
+	log_info("       Left Ctrl or Left GUI - Lay bomb\n");
+#endif
+	log_info("       Left Alt - Ignite bomb\n");
+	log_info("       Left Shift - Jump\n\n");
+	log_info("run with -h to list options\n");
+
+}
+
 void manageTestAI() {
 	static bool doItOnce=true;
 	static bool doItOnce2=true;
@@ -391,16 +661,18 @@ void manageTestAI() {
 			}
 		}
 	} else {
-		if (doItOnce2) {
+		if (cheatMode && doItOnce2) {
 			activeCheatMode();
 			doItOnce2=false;
 		}
 	}
 }
-int exitAtFrame=0;
+int exitAtFrameNumber=0;
+bool exitAtFrame=false;
 int
 main(int argc, char **argv)
 {
+	bool showVersion=false;
 	int c;
 	while (1)
 	{
@@ -421,12 +693,13 @@ main(int argc, char **argv)
 			{"exit", required_argument, 0, '4'},
 			{"tracemask", required_argument, 0, 't'},
 			{"aitest", required_argument, 0, 'a'},
+			{"nomusic", no_argument, 0, 'z'},
 			{0, 0, 0, 0}
 		};
 		/* getopt_long stores the option index here. */
 		int option_index = 0;
 
-		c = getopt_long (argc, argv, "hl:mscv123:4:t:f:o:a:n",
+		c = getopt_long (argc, argv, "hl:mscv123:4:t:f:o:a:nz",
 		                 long_options, &option_index);
 
 		/* Detect the end of the options. */
@@ -444,6 +717,7 @@ main(int argc, char **argv)
 			log_info("  -s, --sex     \t\tSex team mode\n");
 			log_info("  -c, --color     \t\tColor team mode\n");
 			log_info("  -n, --noautofire     \t\tNo autofire for bomb drop\n");
+			log_info("  -z, --nomusic     \t\tNo music\n");
 			log_info("  -v, --version  \t\tDisplay version\n");
 #ifdef DEBUG
 			log_info("Debugging options:\n");
@@ -463,7 +737,7 @@ main(int argc, char **argv)
 			log_info("  -1, --cheat    \t\tActivate L1/L2 pad key for debugging\n");
 			log_info("  -2, --slow    \t\tSlow motion for AI debugging\n");
 			log_info("  -3 <x>, --frame <x>    \tSet frame for randomness debugging\n");
-			log_info("  -4 <x>, --exit <x>    \tExit at frame\n");
+			log_info("  -4 <x>, --exit <x>    \tExit at frame <x> use x = -1 to exit after one game\n");
 			log_info("  -a <x>, --aitest <x>    \tTest <x> AI players\n");
 #endif
 			exit(0);
@@ -483,7 +757,7 @@ main(int argc, char **argv)
 			break;
 		case 'v':
 			log_info("%s %s\n",GAME_NAME,GAME_VERSION);
-			exit(0);
+			showVersion=true;
 			break;
 		case 'l':
 #define NB_LEVELS 8
@@ -508,7 +782,8 @@ main(int argc, char **argv)
 			break;
 		case '4':
 			log_info("-4 option given. Exit at frame %s.\n",optarg);
-			exitAtFrame=atoi(optarg);
+			exitAtFrameNumber=atoi(optarg);
+			exitAtFrame = true;
 			break;
 		case 'a':
 			log_info("-a option given. Test mode for %s AI players.\n",optarg);
@@ -526,12 +801,16 @@ main(int argc, char **argv)
 			log_info("-2 option given. Slow motion for AI debugging.\n");
 			slowMode=true;
 			break;
+		case 'z':
+			music=false;
+			break;
 		default:
 			abort ();
 		}
 	}
-	SDL_Window *window;
-
+#if 0
+	logDebug=fopen ("/tmp/mrboom.log","w");
+#endif
 	/* Enable standard application logging */
 	SDL_LogSetPriority(SDL_LOG_CATEGORY_APPLICATION, SDL_LOG_PRIORITY_INFO);
 	SDL_EventState(SDL_JOYAXISMOTION,SDL_ENABLE);
@@ -539,11 +818,15 @@ main(int argc, char **argv)
 	SDL_EventState(SDL_JOYHATMOTION,SDL_ENABLE);
 	SDL_EventState(SDL_JOYBUTTONDOWN,SDL_ENABLE);
 	SDL_EventState(SDL_JOYBUTTONUP,SDL_ENABLE);
+
 	SDL_JoystickEventState(SDL_ENABLE);
 
-	if (!mrboom_init()) quit(0);
+	if ((exitAtFrame) && (exitAtFrameNumber==-1)) {
+		log_info("No VGA\n");
+		noVGA=SDL_TRUE;
+	}
 
-	if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_JOYSTICK) < 0) {
+	if ((noVGA == SDL_FALSE) && (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_JOYSTICK) < 0)) {
 		log_error("Couldn't initialize SDL: %s\n", SDL_GetError());
 		noVGA = SDL_TRUE;
 		if (SDL_Init(SDL_INIT_JOYSTICK) < 0) {
@@ -552,13 +835,20 @@ main(int argc, char **argv)
 		}
 	}
 
+	if (!mrboom_init()) {
+		log_error("Error in init.\n");
+		quit(1);
+	}
+	if (showVersion) {
+		quit(0);
+	}
+
 	if (noVGA == SDL_FALSE) {
 		/* Create the window and renderer */
 		window = SDL_CreateWindow(GAME_NAME,
 		                          SDL_WINDOWPOS_UNDEFINED,
 		                          SDL_WINDOWPOS_UNDEFINED,
-
-#if 1
+#ifndef FULLSCREEN
 		                          WIDTH*3, HEIGHT*3,
 		                          SDL_WINDOW_RESIZABLE
 #else
@@ -566,10 +856,15 @@ main(int argc, char **argv)
 		                          SDL_WINDOW_FULLSCREEN
 #endif
 		                          );
+
+		windowID = SDL_GetWindowID(window);
+
+
 		if (!window) {
 			log_error("Couldn't set create window: %s\n", SDL_GetError());
 			quit(3);
 		}
+		SDL_SetHint(SDL_HINT_RENDER_SCALE_QUALITY, "best");
 
 		renderer = SDL_CreateRenderer(window, -1, SDL_RENDERER_PRESENTVSYNC);
 		if (!renderer) {
@@ -592,14 +887,35 @@ main(int argc, char **argv)
 #else
 	begin = clock();
 	while (!done) {
+		if (resizeDone) {
+			SDL_SetWindowSize(window, width, height);
+		}
 		loop();
 		fps();
 		if (testAI) manageTestAI();
 		if (slowMode) usleep(100000);
-		if ((exitAtFrame) && (frameNumber()==exitAtFrame)) {
+		if ((exitAtFrame) && (frameNumber()==exitAtFrameNumber)) {
 			log_info("Exit at frame\n");
 			exit(0);
 		}
+		if ((exitAtFrame) && (exitAtFrameNumber==-1)) {
+			static bool first=false;
+			if(isGameActive()==true) {
+				first=true;
+			} else {
+				if (first) {
+					log_info("Exit at frame\n");
+					for (int i=0; i<nb_dyna; i++) {
+						if (victories(i)) {
+							exit(i);
+						}
+
+					}
+					exit(nb_dyna);
+
+				}
+			}
+		}
 	}
 #endif
 
@@ -607,8 +923,7 @@ main(int argc, char **argv)
 		SDL_DestroyRenderer(renderer);
 	}
 	fps();
-
 	log_debug("quit(0)\n");
-	//quit(0);
+	printKeys();
 	return 0;
 }
